(function() {
  window.utu = {
    // queued event calls
    queue: [],
    /**
     * Initializes utu with a token
     * @param  {String} token the utu token for the website
     */
    init: function(token) {
      var browser = this.getBrowserInformation();
      this.token = token;
      this.context = {
        platform: window.location.host,
        platformId: null,
        web: true,
        values: {
          browser: browser[0],
          browserVersion: browser[1],
          userAgent: navigator.userAgent,
        }
      };
      this.setPlatformID()
      this.getGeoInfo();
      this.queueTracking = setInterval(this.checkQueue.bind(this), 1000);
    },

    /**
     * Checks if the queue has queued function calls, if so it will call
     * them if and only if we are not currently trying to fetch the user id
     */
    checkQueue: function() {
      if(this.fetchingID) {
        return
      } else if (this.queue.length > 0) {
        for(var i = 0; i < this.queue.length - 1; i++) {
          try {
            this.queue[i]();
          } catch(e) {
            // if there was an error just remove the object from the queue
            this.queue.splice(i, 1);
          }
        }
        clearInterval(this.queueTracking);
        this.queue = [];
      }
    },

    /**
     * Sends an event to utu
     * @param  {String} event  the name of the event
     * @param  {Object} values the values of the event
     */
    track: function(event, values, callback) {
      // keep reference of the scope
      var self = this;

      var cb = function(event, values) {
        // create a new data object with context and now the event
        var data = self.merge(self.context, {
          event: event
        });

        // update values with the context and the new values of the event
        if (values) {
          data.values = self.merge(data.values, values);
        }

        // send the request to utu
        self.request('https://api.utu.ai/v1/track', {
          data: JSON.stringify(data),
          callback: callback || function() {},
          setHeaders: function(x) {
            x.setRequestHeader('apikey', self.token);
          }
        })
      };

      // if we are still fetching the users information queue the events
      if(self.fetchingID) {
        this.queue.push(function() { cb(event, values); })
      } else {
        cb(event, values);
      }
    },

    /**
     * Create or updates a user in utu
     * @param  {Object} values the values of the user
     */
    identity: function(values, callback) {
      if (!values) {
        throw new Error("identity must have values `utu.identity({ name: 'john doe' })`");
      }
      // keep reference of the scope
      var self = this;

      var cb = function(values) {
        var platform = self.context.platform;
        var platformId = self.context.platformId;
        // create a new data object with context and now the event
        var data = self.merge({ custom: {} }, values);

        // update values with the context and the new values of the event
        data.custom = self.merge(data.custom, self.context.values);

        // send the request to utu
        self.request('https://api.utu.ai/v1/identity/'+platform+'/'+platformId, {
          data: JSON.stringify(data),
          callback: callback || function() {},
          setHeaders: function(x) {
            x.setRequestHeader('apikey', self.token);
          }
        })
      };

      // if we are still fetching the users information queue the events
      if(self.fetchingID) {
        this.queue.push(function() { cb(values); })
      } else {
        cb(values);
      }
    },

    /**
     * Gets information about the browser
     * @return {Array} the browsers information
     */
    getBrowserInformation: function() {
      var N = navigator.appName,
        ua = navigator.userAgent,
        tem;
      var M = ua.match(/(opera|chrome|safari|firefox|msie)\/?\s*(\.?\d+(\.\d+)*)/i);
      if (M && (tem = ua.match(/version\/([\.\d]+)/i)) != null) M[2] = tem[1];
      M = M ? [M[1], M[2]] : [N, navigator.appVersion, '-?'];
      return M;
    },

    /**
     * Creates a UUID for the user
     * @return {String} the uuid
     */
    createUUID: function(cb) {
      var self = this;
      self.fetchingID = true;
      this.request('https://makenu.utu.ai/v1/browser-id', {
        callback: function(response) {
          var resp = JSON.parse(response);
          if(resp && resp.data && resp.data.id) {
            cb(resp.data.id);
            self.fetchingID = false;
          }
        }
      })
    },

    /**
     * Fetches information about the user
     * @return {Object} the information about the user
     */
    getGeoInfo: function() {
      var ctx = this;
      this.request('https://ipinfo.io/json', {
        callback: function(response) {
          if (response) {
            try {
              ctx.context.values = ctx.merge(ctx.context.values, JSON.parse(response));
            } catch (e) {
              console.log(e);
            }
          }
        }
      });
    },

    /**
     * Gets a users id either from a cookie or creates a new one
     * @return {String} the user id
     */
    setPlatformID: function() {
      // try to read the userId from our cookie
      var userId = this.readCookie("utu-uid");
      var self = this;
      // if we don't have a userId create one
      if (!userId) {
        this.createUUID(function(userId) {
          // write and update the cookie
          self.createCookie("utu-uid", userId, 365 * 5);
          self.context.platformId = userId;
        });
      } else {
        // write and update the cookie
        this.createCookie("utu-uid", userId, 365 * 5);
        this.context.platformId = userId;
      }
    },

    /**
     * Merges two objects into a new object
     * @return {Object} the new merged object
     */
    merge: function() {
      return JSON.parse('{' + [].reduce.call(arguments, function(prev, cur) {
          var str = JSON.stringify(cur);
          if (!str || str.slice(0, 1) !== '{' || str.slice(-1) !== '}') {
            throw new Error('Invalid object: ' + str);
          }
          return (prev ? prev + ',' : '') + str.slice(1, -1);
        }, null) +
        '}'
      );
    },

    /**
     * Makes an HTTP request
     * @param  {String} url    the url
     * @param  {Object} config options for the request
     */
    request: function(url, config) {
      try {
        // create the correct type of request for the browser
        var x = new(window.XMLHttpRequest || ActiveXObject)('MSXML2.XMLHTTP.3.0');

        // set the request type
        x.open(config.data ? 'POST' : 'GET', url, 1);

        // set the headers
        x.setRequestHeader('Content-type', 'application/json');

        // set any custom headers
        config.setHeaders && config.setHeaders(x);

        // call the callback when ready
        x.onreadystatechange = function() {
          x.readyState > 3 && config.callback && config.callback(x.responseText, x);
        };

        // send the data
        x.send(config.data)
      } catch (e) {
        window.console && console.log(e);
      }
    },

    /**
     * Creates a cookie
     * @param  {String} name  the name of the cookie
     * @param  {String} value the value for the cookie
     * @param  {Number} days  the number of days the cookie is good for
     */
    createCookie: function(name, value, days) {
      if (days) {
        var date = new Date();
        date.setTime(date.getTime() + (days * 24 * 60 * 60 * 1000));
        var expires = "; expires=" + date.toGMTString();
      } else var expires = "";
      document.cookie = name + "=" + value + expires +"; domain=." + location.hostname.split('.').reverse()[1] + "." + location.hostname.split('.').reverse()[0] + "; path=/";
    },

    /**
     * Reads a cookie
     * @param  {String} name the name of the cookie
     * @return {String}      the value of the cookie
     */
    readCookie: function(name) {
      var nameEQ = name + "=";
      var ca = document.cookie.split(';');
      for (var i = 0; i < ca.length; i++) {
        var c = ca[i];
        while (c.charAt(0) == ' ') c = c.substring(1, c.length);
        if (c.indexOf(nameEQ) == 0) return c.substring(nameEQ.length, c.length);
      }
      return null;
    },
  };
})()
